from dataclasses import dataclass, field
from typing import Any, Optional, Protocol, Union

JSONType = Union[dict[str, "JSONType"], list["JSONType"], str, int, float, bool, None]


@dataclass
class PipelineIteratedValue:
    """Class to hold the iterated value of a pipeline"""

    value: Any
    index: int


class PipelineFunction(Protocol):
    """Generic Function signature for every MCPF function

    This type defines the expected signature of every pipeline function.

    """

    def __call__(
        data: dict[str, Any],
        meta: Optional[dict[str, Any]] = None,
        has_iterated_value: bool = False,
        iterated_value: Optional[Any] = None,
        iterated_index: int = -1,
        arg: Optional[dict[str, JSONType]] = None,
    ) -> dict[str, Any]:
        """Generic Function signature for every MCPF function

        Generic Function signature for every MCPF function.

        Only the data parameter is mandatory. The rest of the parameters need not
        be declared in the function signature if you don't need them. Each optional
        parameter will be injected by the pipeline engine, so you can selectively
        query the information you actually need.

        Args:
            data (dict[str, Any]): The main data dictionary to process.
                It is up to the caller to define any conventions to pass data down
                the pipeline. This parameter is mandatory.
            meta (Optional[dict[str, Any]]): Contains metadata derived from the pipeline
                configuration. You can change metadata in the function, and the changed
                metadata will be passed to the next function in the pipeline.
            has_iterated_value (bool): If this function is to be used within a loop or
                other iterable construct, you can query whether a value for the current
                iteration step is still available.
            iterated_value (Optional[Any]): The current iterated value, if any.
            iterated_index (int): The index of the current iterated value or -1 if not
                within an iteration step.
            arg (Optional[dict[str, JSONType]]): Contains a mapping of names to
                argument names to be used as keys in the data dictionary. The arg
                dictionary is extracted from the meta dictionary.

        Returns:
            dict[str, Any]: The potentially modified data dictionary ready to be passed
                to the next function in the pipeline. If None, the `data` argument
                will be implicitly returned. Even if you return None, all changes
                applied to the data dictionary will still be propagated to the next
                function in the pipeline.

                If you want to prevent changes to the data dictionary from being
                propagated, you can make a deep copy of the data dictionary and
                only change the copy within the function.
        """
        ...


@dataclass
class DatabaseConfig:
    type: str
    url: str
    user: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None
    org: Optional[str] = None
    bucket: Optional[str] = None


@dataclass
class PipelineConfig:
    input_path: str
    output_path: str
    entry_point: str
    imports: list[str]
    pipelines: list[dict[str, list[dict[str, str]]]]
    input_file_name: str = field(default=None)
    tmp_paths: list[str] = field(default=None)
    pipeline_extension: list[dict[str, list[dict[str, str]]]] = field(default=None)
    further_configuration: Optional[list[dict[str, str]]] = field(default=None)
    database_configs: Optional[list[DatabaseConfig]] = field(default=None)



"""
This module contains decorators for pipeline functions.
"""
import functools
import inspect


def with_default_arguments(default_arguments: dict[str, str]):
    """
    This decorator allows you to set default arguments for a pipeline function
    and will merge them with the arguments provided in the function call.
    Arguments provided by parameter 'arg' will override the default ones."

    ```python
    @with_default_arguments({
        "input": constants.DEFAULT_IO_DATA_LABEL,
        "output": constants.DEFAULT_IO_DATA_LABEL,
    })
    def my_pipeline_function(data: dict[str, str], arg: dict[str, str]):
        # Your function implementation here
        ...
    ```

    is the equivalent of:
    ```python
    def my_pipeline_function(data: dict[str, Any], meta: dict[str, Any]):
        # default_arguments_values
        arg = {
            "input": constants.DEFAULT_IO_DATA_LABEL,
            "output": constants.DEFAULT_IO_DATA_LABEL,
        }
        # merging default values with current argument values
        if meta[constants.ARGUMENTS]:
            arg = arg | meta[constants.ARGUMENTS]

        # Your function implementation here
        ...
    ```

    """
    if not isinstance(default_arguments, dict):
        raise TypeError("default_arguments must be a dictionary")

    def my_decorator(func):
        sig = inspect.signature(func)
        apply = "arg" in sig.parameters

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if apply:
                arg = default_arguments.copy()
                passed_arg = kwargs.get("arg", None)
                if passed_arg is not None:
                    arg |= passed_arg

                kwargs["arg"] = arg
            return func(*args, **kwargs)

        return wrapped

    return my_decorator



#! /usr/bin/env python3
# TODO
# config levels
# more and embedded loops (list of loop list and iterator list)

import dataclasses
import functools
import importlib
import json
import os.path
from pathlib import Path
from types import ModuleType
from typing import Any, Optional

import risc_lasagna as lasagna
import risc_lasagna.layer as layer
from toolz import pipe

import mcpf_core.core.constants as constants
import mcpf_core.core.routines as routines
import mcpf_core.core.singletons as singletons
from mcpf_core.core import arginjector
from mcpf_core.core.types import JSONType, PipelineConfig, PipelineFunction

dept_of_nested_loops = 0
real_dept_of_loop_kernel_hierarchy = 0
current_max_dept_of_nested_loops = 0
loop_kernel_pipelines: list[
    tuple[str, list[PipelineFunction]]
] = []  # it contains the child pipeline of loops in the order of execution

dict_of_called_pipeline_element: dict[str, int] = {}


def load_pipeline_config(args: Optional[list[str]]) -> PipelineConfig:
    """
    Loads the pipeline configuration using the "lasagna" package to parse YAML configuration files and compose a single code pipeline configuration.

    Args:
        args (Optional[list[str]]): list of command-line arguments. If None or empty, default configuration is used. Otherwise, the list contains paths to YAML configuration files.

    Returns:
        PipelineConfig: A dataclass containing the merged pipeline configuration.

    Raises:
        FileNotFoundError: If any of the provided YAML configuration files do not exist.
        NotImplementedError: If the configuration is missing required fields such as 'imports', 'pipelines', or 'entry_point'.
    """
    # set the environment variable with a prefix
    if not args or len(args) == 0:
        return lasagna.build(
            PipelineConfig,
            [
                layer.DataClassDefaultLayer(
                    PipelineConfig(
                        ".", ".", "default_p", [], [{"default_p": [{"version": "~"}, {"help": "~"}]}], "", []
                    )
                ),
            ],
        )
    else:
        extensions = []
        imports = dict[str, str]()
        converted_list = map(Path, args)
        yaml_layers = list(map(layer.YamlLayer, converted_list))
        for l in yaml_layers[:-1]:
            if constants.ARG_KEYWORD_PIPELINE_EXT in l.map:
                extensions.append(l.map[constants.ARG_KEYWORD_PIPELINE_EXT])
        for l in yaml_layers:
            if constants.ARG_KEYWORD_IMPORTS in l.map:
                imports |= {imp: imp for imp in l.map[constants.ARG_KEYWORD_IMPORTS]}
        l_config = lasagna.build(
            PipelineConfig,
            [
                *yaml_layers,
                layer.DataClassDefaultLayer(
                    PipelineConfig(
                        ".", ".", "default_p", [], [{"default_p": [{"version": "~"}, {"help": "~"}]}], "", []
                    )
                ),
            ],
        )
        l_config.imports = list[str](imports.values())
        first_element = True
        for sub_pipeline in l_config.pipelines:
            if first_element:
                first_element = False
            else:
                for key in sub_pipeline:
                    l_config.pipelines[0][key] = sub_pipeline[key]

        del l_config.pipelines[1:]

        if len(extensions) > 0:
            for extension in reversed(extensions):
                if len(extension) > 0:
                    for extension_element in extension:
                        for pipeline_element in extension_element:
                            l_config.pipelines[0][pipeline_element] = extension_element[pipeline_element]
        return l_config


def increment_dept_of_nested_loops():
    """
    This function keeps track the dept of the nested loops.
    """
    global current_max_dept_of_nested_loops
    global dept_of_nested_loops
    global real_dept_of_loop_kernel_hierarchy
    dept_of_nested_loops += 1
    real_dept_of_loop_kernel_hierarchy += 1
    if real_dept_of_loop_kernel_hierarchy > current_max_dept_of_nested_loops:
        current_max_dept_of_nested_loops = real_dept_of_loop_kernel_hierarchy


def init_iterator(data: dict[str, Any]) -> dict[str, Any]:
    """
    Each execution of every loop kernel starts with this function. It initializes the current value of the
    loop iterator.
    """
    global dept_of_nested_loops
    loop_singleton = singletons.LoopIterators()
    loop_singleton.init_current_iterator(dept_of_nested_loops - 1)
    return data


def loop_interpreter(data: dict[str, Any]) -> dict[str, Any]:
    """
    This function is used to implement a loop in the code pipeline.
    """
    global dept_of_nested_loops  # Handling iterator list
    global current_max_dept_of_nested_loops
    global loop_kernel_pipelines
    global real_dept_of_loop_kernel_hierarchy  # Handling loop kernels
    loop_singleton = singletons.LoopIterators()
    param_singleton = singletons.Arguments()
    if len(loop_kernel_pipelines) > 0:
        increment_dept_of_nested_loops()
        kernel_entry = loop_kernel_pipelines[real_dept_of_loop_kernel_hierarchy - 1]
        if kernel_entry[0] not in dict_of_called_pipeline_element:
            dict_of_called_pipeline_element[kernel_entry[0]] = real_dept_of_loop_kernel_hierarchy
        else:
            real_dept_of_loop_kernel_hierarchy = dict_of_called_pipeline_element[kernel_entry[0]]
            param_singleton.overwrite_current_depth(real_dept_of_loop_kernel_hierarchy - 1)
        kernel = loop_kernel_pipelines[real_dept_of_loop_kernel_hierarchy - 1][1]
        param_singleton.replace_current_argument_lists()
        try:
            while loop_singleton.size_of_an_iterator_list(dept_of_nested_loops - 1) > 0:
                data = pipe(data, *kernel)
                param_singleton.renew_loop_argument_lists()
        finally:
            param_singleton.restore_last_buffered_argument_list()
            routines.pop_loop_iterator()  # it is needed just in that case if no one used up the iterator
            loop_singleton.remove_an_iterator_list(dept_of_nested_loops - 1)
            if dept_of_nested_loops == real_dept_of_loop_kernel_hierarchy:
                real_dept_of_loop_kernel_hierarchy -= 1
            dept_of_nested_loops -= 1
            if dept_of_nested_loops == 0:
                del loop_kernel_pipelines[:current_max_dept_of_nested_loops]
                current_max_dept_of_nested_loops = 0
                dict_of_called_pipeline_element.clear()
    return data


def _wrap_pipeline_function_injector(func):
    """
    This function is used to inject the optional arguments into the function itself.
    That way you need less boilerplate code in the function itself.
    """

    arginjector.assert_parameters(func)

    @functools.wraps(func)
    def wrapper(data: dict[str, Any], *args, **kwargs):
        return arginjector.inject_arguments(func, data, *args, **kwargs)

    return wrapper


def validate_pipeline(
    config: PipelineConfig,
    current_pipeline: str,
    modules: dict[str, ModuleType],
    current_param_lists: list[dict[str, JSONType]],
    param_lists_of_loops: list[list[Any]],
    evaluated_sub_pipelines=None,  # avoiding to validate recursive calls
    is_recursive: bool = False,
) -> list[PipelineFunction]:
    """
    This function compose the code pipeline and pipeline of the loop kernels according the given yaml configuration,
    stores the function arguments given in the yaml configuration in a structured way and
    validate each pipeline (e.g.: it checks the existence of the given functions in the enumerated python modules).
    """
    global loop_kernel_pipelines
    # preparing available building blocks
    # this_module = sys.modules[__name__]
    if evaluated_sub_pipelines is None:
        evaluated_sub_pipelines = []
    available_functions: dict[str, list[PipelineFunction]] = {}
    for module in modules:
        function_list = dir(modules[module])
        available_functions[module] = function_list

    pipeline: list[PipelineFunction] = []
    for func_const in config.pipelines[0][current_pipeline]:
        for func in func_const:
            if func[: len(constants.ARG_KEYWORD_LOOP)] == constants.ARG_KEYWORD_LOOP:
                if not is_recursive:
                    if func_const[func] in evaluated_sub_pipelines:
                        is_recursive = True
                    child_pipeline = [init_iterator]
                    kernel_arguments_lists = []
                    param_lists_of_loops.append(kernel_arguments_lists)
                    loop_kernel_pipelines.append((func_const[func], child_pipeline))
                    pipeline.append(loop_interpreter)
                    child_pipeline.extend(
                        validate_pipeline(
                            config,
                            func_const[func],
                            modules,
                            kernel_arguments_lists,
                            param_lists_of_loops,
                            evaluated_sub_pipelines + [func_const[func]],
                            is_recursive,
                        )
                    )
                    # f = getattr(this_module, 'init_iterator')
            elif func in config.pipelines[0]:
                if func not in evaluated_sub_pipelines:
                    evaluated_sub_pipelines.append(func)
                    sub_pipeline = validate_pipeline(
                        config,
                        func,
                        modules,
                        current_param_lists,
                        param_lists_of_loops,
                        evaluated_sub_pipelines + [func],
                        is_recursive,
                    )
                    pipeline.extend(sub_pipeline.copy())
            else:
                not_found = True
                for module in available_functions:
                    if func in available_functions[module]:
                        f: PipelineFunction = getattr(modules[module], func)
                        pipeline.append(_wrap_pipeline_function_injector(f))
                        current_param_lists.append(func_const)
                        not_found = False
                        break
                if not_found:
                    raise NotImplementedError("Error: None of the imported modules contain the function " + func + "!")
    return pipeline


def run_pipeline(
    config: PipelineConfig, pipeline: list[PipelineFunction], current_param_lists: list[dict[str, str]]
) -> dict[str, Any]:
    """
    It initializes the "meta" and execute the code pipeline composed from the given configuration.
    """
    config.tmp_paths.insert(0, config.input_path)
    config.tmp_paths.append(config.output_path)
    tmp_absolut_paths = []
    for path in config.tmp_paths:
        tmp_absolut_paths.append(os.path.abspath(path))
    meta = {constants.TMP_PATH_INDEX: 0}
    meta[constants.TMP_PATHS] = tmp_absolut_paths
    if config.further_configuration:
        for key_value_pair in config.further_configuration:
            for key in key_value_pair:
                meta[key] = key_value_pair[key]
    if config.database_configs:
        db_configs = []
        for config_element in config.database_configs:
            config_dict = {}
            for f in dataclasses.fields(config_element):
                config_dict[f.name] = getattr(config_element, f.name)
            db_configs.append(config_dict)
        meta[constants.DB_CONFIG] = db_configs
    json_meta = json.dumps(meta)
    data = {constants.ARG_KEYWORD_META: json_meta}
    retval: dict[str, Any] = pipe(data, *pipeline)
    return retval


def run(*args: str) -> dict[str, Any]:
    """
    Run pipeline programmatically as if called from command line
    """
    singletons.LoopIterators.reset()
    singletons.Arguments.reset()
    for arg in args:
        if not os.path.isfile(arg):
            raise FileNotFoundError("Error: config file " + arg + " does not exist.")
    c = load_pipeline_config(args)
    if not c.imports or not c.pipelines or not c.entry_point:
        raise NotImplementedError("Error: Missing 'imports', 'pipelines' or 'entry_point' in the config file.")
    else:
        imported_modules: dict[str, ModuleType] = {}
        for module_name in c.imports:
            i = importlib.import_module(module_name)
            imported_modules[module_name] = i

        param_lists: list[dict[str, str]] = []
        arguments_of_sub_pipelines: list[list[Any]] = []
        pipeline_string = validate_pipeline(c, c.entry_point, imported_modules, param_lists, arguments_of_sub_pipelines)
        p = singletons.Arguments()
        p.current_param_lists = param_lists
        p.param_lists_of_loops = arguments_of_sub_pipelines
        return run_pipeline(c, pipeline_string, param_lists)



"""
time.py

This module provides utility functions for working with timestamps.

Functions:
    - convert_second_to_ms(timestamp_in_seconds: int) -> int:
        Converts a timestamp from seconds to milliseconds.
    - convert_second_to_us(timestamp_in_seconds: int) -> int:
        Converts a timestamp from seconds to microseconds.
"""


def convert_second_to_ms(timestamp_in_seconds: int):
    """
    Converts a timestamp from seconds to milliseconds.

    Args:
        timestamp_in_seconds (int): The timestamp in seconds.

    Returns:
        int: The timestamp converted to milliseconds.
    """
    return timestamp_in_seconds * 1000


def convert_second_to_us(timestamp_in_seconds: int):
    """
    Converts a timestamp from seconds to microseconds.

    Args:
        timestamp_in_seconds (int): The timestamp in seconds.

    Returns:
        int: The timestamp converted to microseconds.
    """
    return timestamp_in_seconds * 1000 * 1000



"""
Performs resolution of arguments for injection.
"""

import inspect
from typing import Any

import mcpf_core.core.routines as routines
from mcpf_core.core import constants, singletons
from mcpf_core.core.types import PipelineFunction

_VALID_PARAMETERS = {
    "data",
    "meta",
    "arg",
    "has_iterated_value",
    "iterated_value",
    "iterated_index",
}


def assert_parameters(func: PipelineFunction) -> None:
    """
    Asserts that all specified parameters are valid for argument injection.
    """
    sig = inspect.signature(func)
    invalid_parameters = [p for p in sig.parameters if p not in _VALID_PARAMETERS]
    if invalid_parameters:
        raise KeyError(
            f"Invalid parameters in function '{func.__name__}': {invalid_parameters}. Valid parameters are: {_VALID_PARAMETERS}"
        )


def inject_arguments(func: PipelineFunction, data: dict[str, Any], **kwargs):
    prepared_args = [data]
    prepared_kwargs = dict(**kwargs)

    sig = inspect.signature(func)

    needs_meta = "meta" in sig.parameters or "arg" in sig.parameters

    if needs_meta:
        meta = routines._get_meta_data_for_function(data, func)

    meta_specified = False
    if "meta" in sig.parameters:
        prepared_kwargs["meta"] = meta
        meta_specified = True

    if "has_iterated_value" in sig.parameters:
        prepared_kwargs["has_iterated_value"] = routines.is_iterated_value_available()

    if "iterated_value" in sig.parameters or "iterated_index" in sig.parameters:
        iteratedValue = singletons.LoopIterators().pop_iterated_value()
        if "iterated_value" in sig.parameters:
            prepared_kwargs["iterated_value"] = iteratedValue.value if iteratedValue else None
        if "iterated_index" in sig.parameters:
            prepared_kwargs["iterated_index"] = iteratedValue.index if iteratedValue else -1

    if "arg" in sig.parameters:
        arg = kwargs.get("arg", None)
        if constants.ARG_KEYWORD_ARGUMENTS in meta and isinstance(meta[constants.ARG_KEYWORD_ARGUMENTS], dict):
            arg = arg or {}
            arg = arg | meta[constants.ARG_KEYWORD_ARGUMENTS]
        prepared_kwargs["arg"] = arg

    result = func(*prepared_args, **prepared_kwargs)

    if result is not None and isinstance(result, dict):
        data = result

    if meta_specified:
        routines.set_meta_in_data(data, meta)

    return data



from pathlib import Path

ARG_KEYWORD_IMPORTS = "imports"
ARG_KEYWORD_LOOP = "loop"
ARG_KEYWORD_META = "meta"
ARG_KEYWORD_ARGUMENTS = "arguments"
ARG_KEYWORD_PIPELINE_EXT = "pipeline_extension"

DEFAULT_IO_DATA_LABEL = ""
INPUT_PATH = "input_path"
OUTPUT_PATH = "output_path"
TMP_PATHS = "tmp_paths"
TMP_PATH_INDEX = "tmp_path_index"
DB_CONFIG = "database_configs"

PROJECT_PATH = Path(__file__).parents[1]
CSV_NAME_PATH = PROJECT_PATH / "mcp_use_case_testing_duckdb" / "testing_csv.csv"



import sys

from mcpf_core.core import runtime

if len(sys.argv) < 2:
    print(
        """Usage: python -m mcpf_core.run [extension_pipeline_config.yaml...] pipeline_config.yaml
    
extension_pipeline_config.yaml: Path to the extension pipeline configuration file.
          Note that extension pipeline files are processed from right to left.
pipeline_config.yaml: Path to the pipeline configuration file.
"""
    )
    sys.exit(1)

runtime.run(*sys.argv[1:])



import copy
from typing import Any, Iterable, Optional

from mcpf_core.core.types import JSONType, PipelineIteratedValue


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]

    def reset(cls):
        """
        This function clears the singleton instances of the given class.
        Args:
            cls: The class whose singleton instance are to be cleared.
        """
        cls._instances = {}


class LoopIterators(metaclass=SingletonMeta):
    """
    This is a singleton class to implicitly providing the loop iterators.
    """

    def __init__(self):
        self.listsOfLoopIterators: list[PipelineIteratedValue] = []
        self.iterator: PipelineIteratedValue = None

    def register_new_iterator_list(self, iterator_list: Iterable[Any], deep_copy: bool = False):
        """
        This function is called, if the element of the list will be used as iterator values in a subsequent loop.
        Args:
            iterator_list:  the list of the iterator values.
            deep_copy:      if it is true the value of the elements of the list will be copied as well
                            (not only their references).

        """
        if deep_copy:
            iterator_list = copy.deepcopy(iterator_list)

        self.listsOfLoopIterators.append([PipelineIteratedValue(v, i) for i, v in enumerate(iterator_list)])

    def size_of_an_iterator_list(self, index: int) -> int:
        """
        It returns the size of the corresponding registered list of loop iterator values.
        Args:
            index: This is the index of the list (in the order of registration).

        Returns: length of the corresponding list.

        """
        if len(self.listsOfLoopIterators) < index:
            return 0
        else:
            # Corner case: if loop was never prepared, we pretend an empty list
            if len(self.listsOfLoopIterators) <= index:
                return 0
            return len(self.listsOfLoopIterators[index])

    def remove_an_iterator_list(self, index: int):
        """
        It removes the corresponding registered list of loop iterator values.
        Args:
            index: This is the index of the list (in the order of registration).
        """
        if len(self.listsOfLoopIterators) > index:
            del self.listsOfLoopIterators[index]

    def init_current_iterator(self, index: int):
        """
        It assigns the next value for variable self.iterator
        from the corresponding registered list of loop iterator values.
        Args:
            index: This is the index of the list (in the order of registration).
        """
        if len(self.listsOfLoopIterators) > index and len(self.listsOfLoopIterators[index]) > 0:
            self.iterator = self.listsOfLoopIterators[index].pop(0)

    def is_iterated_value_available(self) -> bool:
        """
        It returns true if the iterated value is still available and can be consumed with `pop_iterated_value`.
        """
        return self.iterator is not None

    def pop_iterated_value(self) -> Optional[PipelineIteratedValue]:
        """
        It returns the value of the variable self.iterator. The self.iterator is set to 'None'.

        Note: If `pop_iterator` returns None, the function is not called within a loop.
        """
        iterator = self.iterator
        self.iterator = None
        return iterator

    def pop_iterator(self) -> Any:
        """
        It returns the value of the variable self.iterator.value. The self.iterator is set to 'None'.

        Note: Don't use this function. You cannot discriminate against the iterator not being available
        and the actual iterated value to be falsy. Use `pop_iterated_value` instead.
        """
        iterator = self.pop_iterated_value()
        return iterator.value if iterator else None


class Arguments(metaclass=SingletonMeta):
    """
    This is a singleton class to implicitly providing function arguments given in the yaml configuration.
    """

    def __init__(self):
        self.current_param_lists = None
        self.buffered_param_lists = []
        self.buffered_dept_at_recursion = []
        self.param_lists_of_loops = None
        self.current_dept = 0
        self.max_dept = 0

    def get_upcoming_arguments(self, caller_name: str) -> dict[str, JSONType] | None:
        """
        It provides the upcoming yaml arguments for the function whose name is given as argument.
        Args:
            caller_name: name of the function, whose upcoming arguments will be return.

        Returns:
            A dictionary were the keys are the arguments name specified in the yaml configuration and the values are
            the values of the given arguments.
        """
        ret_val = None
        while self.current_param_lists and (not ret_val or caller_name not in ret_val):
            ret_val = self.current_param_lists.pop(0)
        if caller_name in ret_val:
            if ret_val[caller_name] and type(ret_val[caller_name]) is list:
                return ret_val[caller_name][0]
            else:
                return None
        else:
            return None

    def overwrite_current_depth(self, new_value: int):
        # This function is called only in case of recursion.
        # In this case the parameter queue (as well as the queue the pipeline of the loop kernel)
        # contains an extra entry (the data of the entry point of the recursion or in other words
        # the first recursively called loop kernel), that is why max_dept must be greater with 1.
        if self.current_dept + 1 > self.max_dept:
            self.max_dept = self.current_dept + 1
        self.buffered_dept_at_recursion.append(self.current_dept)
        self.current_dept = new_value

    def replace_current_argument_lists(self):
        """
        The arguments for loop kernel are stored in independent lists.
        If the execution a loop kernel is started, then parent list
        (belonging to the main list or a containing loop kernel) will be buffered and the subsequent
        list will be activated.
        """
        self.buffered_param_lists.append(self.current_param_lists)
        self.current_param_lists = self.param_lists_of_loops[self.current_dept].copy()
        self.current_dept += 1
        if self.current_dept > self.max_dept:
            self.max_dept = self.current_dept

    def renew_loop_argument_lists(self):
        """
        At each execution of a loop kernel, the list of arguments of its functions must be reinitialized.
        This is done by this function.
        """
        if self.current_dept - 1 > 0:
            self.current_param_lists = self.param_lists_of_loops[self.current_dept - 1].copy()
        else:
            self.current_param_lists = self.param_lists_of_loops[0].copy()

    def restore_last_buffered_argument_list(self):
        """
        If the execution returns to the main pipeline or the pipeline of the containing loop kernel, its buffered
        parameter list is re-activated.
        """
        self.current_param_lists = self.buffered_param_lists.pop()
        if len(self.buffered_dept_at_recursion) > 0:
            self.current_dept = self.buffered_dept_at_recursion.pop()
        else:
            self.current_dept -= 1
        if self.current_dept == 0:
            del self.param_lists_of_loops[: self.max_dept]
            self.max_dept = 0



import inspect
import json
import sys
from typing import Any, Optional, Tuple, Union

import mcpf_core.core.constants as constants
import mcpf_core.core.singletons as singleton
from mcpf_core.core.types import PipelineFunction


def get_meta_data(data: dict[str, Any], func: Optional[PipelineFunction] = None) -> dict[str, Any]:
    """
    It extracts the json string "meta" from "data" and adds the current yaml arguments of the caller to "meta".
    Returns:
        A dictionary, which contains the content of the "meta" as well as the current arguments of the caller function.
    """
    caller_func = inspect.getframeinfo(sys._getframe(1))[2]
    return _get_meta_data_for_function(data, caller_func)


def _get_meta_data_for_function(data: dict[str, Any], func: Union[PipelineFunction, str]) -> dict[str, Any]:
    """
    Extracts and updates metadata for a given function.

    This function retrieves the JSON string "meta" from the provided `data` dictionary,
    parses it into a dictionary, and adds the current YAML arguments of the caller
    function to the metadata. The caller function is identified either by its name
    (if `func` is a string) or by the `__name__` attribute (if `func` is a `PipelineFunction`).

    Note: This function is intended for internal use and should not be called directly.
    Use `get_meta_data` function instead from within pipeline functions.

    Args:
        data (dict[str, Any]): A dictionary the metadata is derived from.
        func (Union[PipelineFunction, str]): The function or its name for which
            arguments are to be injected into the metadata.

    Returns:
        dict[str, Any]: A dictionary containing the parsed metadata along with the
        current arguments of the caller function.

    Raises:
        ValueError: If no metadata could be derived from the `data` dictionary.
    """
    if constants.ARG_KEYWORD_META not in data:
        raise ValueError("Error! No metadata in the arguments")
    json_meta = data[constants.ARG_KEYWORD_META]
    meta = json.loads(json_meta)
    caller_func = func if type(func) == str else func.__name__
    param_singleton = singleton.Arguments()
    meta[constants.ARG_KEYWORD_ARGUMENTS] = param_singleton.get_upcoming_arguments(caller_func)
    return meta


def set_meta_in_data(data: dict[str, Any], meta: dict[str, Any]):
    """
    It packs back the meta information given in terms of a dictionary into json string and stores it in "data".
    Args:
        data:
        meta:

    Returns:
        data
    """
    del meta[constants.ARG_KEYWORD_ARGUMENTS]
    json_meta = json.dumps(meta)
    data[constants.ARG_KEYWORD_META] = json_meta


def register_loop_iterator_list(iterator_list: list, deep_copy: bool = False):
    """
    It registers a list of loop iterator values.
    Args:
        iterator_list:
        deep_copy:
    """
    loop_singleton = singleton.LoopIterators()
    loop_singleton.register_new_iterator_list(iterator_list, deep_copy)


def is_iterated_value_available() -> bool:
    """
    Returns true if the iterated value is still available and can be consumed with `pop_loop_iterator`.
    """
    loop_singleton = singleton.LoopIterators()
    return loop_singleton.is_iterated_value_available()


def pop_loop_iterator() -> Any:
    """
    It returns with the current iterator value, if it wasn't returned yet.

    Note: Use `pop_iterated_value` instead. It performs the same function,
    but its name better conveys the function's meaning.

    Returns:
        object: the current iterator value or None.
    """
    return pop_iterated_value()


def pop_iterated_value() -> Any:
    """
    It returns the current iterated value.

    If there was none, or the iterated value was already consumed by previously calling
    `pop_iterated_value` (either by the calling function or a preceding function in the same
    pipeline), it returns None.

    Check with `is_iterated_value_available` if the value was already consumed,
    or if the value is available but actually `None`.

    Returns:
        the current iterated value or None, depending on the state of `is_iterated_value_available`.
    """
    value, _ = pop_iterated_value_and_index()
    return value


def pop_iterated_value_and_index() -> Tuple[Any, int]:
    """
    It returns the current iterated value and the current loop index.

    If there was none, or the iterated value was already consumed by previously calling
    `pop_iterated_value` (either by the calling function or a preceding function in the same
    pipeline), it returns None.

    Check with `is_iterated_value_available` if the value was already consumed,
    or if the value is available but actually `None`.

    Returns:
        a tuple of the current iterated value and index, or `(None, -1)`,
        depending on the state of `is_iterated_value_available`.
    """
    loop_singleton = singleton.LoopIterators()
    iterated_value = loop_singleton.pop_iterated_value()
    return (iterated_value.value, iterated_value.index) if iterated_value else (None, -1)


def get_current_input_dir(meta: dict[str, Any]) -> str:
    """
    It returns with current input directory.
    Args:
        meta: the required information will be extracted from "meta".
    """
    index_of_dirs = meta[constants.TMP_PATH_INDEX]
    input_dir = meta[constants.TMP_PATHS][index_of_dirs]
    return input_dir


def get_current_tmp_dir(meta: dict[str, Any]) -> str:
    """
    It returns with current tmp/output directory
    Args:
        meta: the required information will be extracted from "meta".
    """
    index_of_dirs = next_tmp_dir_index(meta)
    output_dir = meta[constants.TMP_PATHS][index_of_dirs]
    return output_dir


def get_final_output_dir(meta: dict[str, Any]) -> str:
    """
    It returns with directory to which the final output should be stored.
    Args:
        meta: the required information will be extracted from "meta".
    """
    index_of_dirs = next_tmp_dir_index(meta)
    output_dir = meta[constants.TMP_PATHS][-1]
    return output_dir


def next_tmp_dir_index(meta: dict[str, Any]) -> int:
    """
    It returns with the index of current tmp/output directory.
    Args:
        meta: the required information will be extracted from "meta".
    """
    index_of_dirs = meta[constants.TMP_PATH_INDEX]
    if len(meta[constants.TMP_PATHS]) - 1 > index_of_dirs:
        index_of_dirs += 1
    return index_of_dirs


def set_current_output_dir_to_input_dir(meta: dict[str, Any]):
    """
    It sets the current tmp directory to the current input directory and the upcoming tmp directory to the current one.
    """
    meta[constants.TMP_PATH_INDEX] = next_tmp_dir_index(meta)


def get_db_config(meta: dict[str, Any], db_type: str) -> dict[str, Any] | None:
    """
    It returns with "database_configs" part of the current yaml configuration.
    Args:
        meta:
        db_type:

    Returns:
        A dictionary, which contains how to access to the database, whose type corresponds to the given db type.
    """
    if constants.DB_CONFIG not in meta or len(meta[constants.DB_CONFIG]) == 0:
        return None
    ret_val_dict: dict[str, Any] = meta[constants.DB_CONFIG][0]
    index = 1
    while ret_val_dict["type"] != db_type and index < len(meta[constants.DB_CONFIG]):
        ret_val_dict = meta[constants.DB_CONFIG][index]
        index += 1
    if ret_val_dict["type"] == db_type:
        return ret_val_dict
    else:
        return None
